/*
	Word Twiddle
	JavaScript game engine
	Version 0.9
	Copyright 2011 by Andrew Krywaniuk

	Using iframes to load the data because it works in standalone mode, whereas Ajax requests wouldn't.
	However, Ajax is used to load rounds from the server.
		
	License is hereby granted to freely distribute this game for personal and non-commercial use.
	For other uses, please contact the author.
*/


// total_word_lists is an array of arrays for each word length (using 1-based indexing).
var total_word_lists = [ null, null, null, null, null, null, null, null];

// word map is a flat list of all words indexed by hash value.
var total_word_map = new Object();

var g_dictionary_name = null;
var dictionary_load_op = null;


// reset_dictionary - clear the current dictionary.
function reset_dictionary()
{
	total_word_lists = [ null, null, null, null, null, null, null, null];
	total_word_map = new Object();
}


// letter vals - used in the hash function.
var letter_vals = [
	0x00100101, 0x000d0000, 0x00800c00, 0x00300010, 0x02004010, 0x000400b0, 0x00006000,  // abcdefg
	0x00000201, 0x0000d010, 0x000000c0, 0x00030100, 0x00040010, 0x00000130, 0x00005010,  // hijklmn
	0x00301010, 0x00000e04, 0x00000120, 0x000030c0, 0x00404001, 0x00080120, 0x00015030,  // opqrstu
	0x00060020, 0x00040d20, 0x00080506, 0x00003107, 0x0000010c  // vwxyz
	];
	

// hash_string - create a collision-resistant hash of a word.
function hash_string(str)
{
	var n = str.length;
	var i;
	var ch;
	var hash = 0;
	
	for (i=0; i<n; i++)
	{
		ch = str.charCodeAt(i);
		idx = ch - "a".charCodeAt(0);
		
		// Check for illegal characters in string.
		if (idx < 0 || idx >= 26) return 0;
		
		hash += letter_vals[idx];
	}
	
	return hash;
}


// scramble_word - create a random anagram of a word.
function scramble_word(wd)
{
	var aChars = wd.toUpperCase().split("");
	var new_wd = "";
	
	for (var len = aChars.length; len>0; len--)
	{
		var x = gen_rand(len);
		new_wd += aChars.splice(x,1);
	}
	
	return new_wd;
}


// load_wt_data - load the entire dictionary of length N from a word list that has
// been pre-sorted by word length.
function load_wt_data(word_len)
{
	var frm_name = "w" + word_len;
	var ofrm = document.getElementById(frm_name);
	var data = ofrm.contentDocument.body.innerHTML
	var len = data.length;
	
	// Create a new word list for this word length.
	var aWords = []; 
	
	
	// strip the <pre> & </pre>
	data = data.substr(5, len - 11);
	
	var i;
	var j;
	
	var a_lines = data.split("\n");
	var n = a_lines.length;
	
	for (i=0; i<n; i++)
	{
		var a_words = a_lines[i].split(" ");
		var m = a_words.length;
		
		for (j=0; j<m; j++)
		{
			var wd = a_words[j];
			if (wd.length == 0) continue;
			
			aWords.push(wd);
			
			var hash = hash_string(wd);
			
			if (total_word_map[hash])
			{
				total_word_map[hash].push(wd);
			}
			else
			{
				total_word_map[hash] = [wd];
			}
		}
	}
	
	// Store this word list in RAM.
	total_word_lists[word_len] = aWords;
	
	on_dictionary_frame_loaded(frm_name);
}


// load_unsorted_wt_data - load word data from a dictionary that has not been
// pre-sorted by word length.
function load_unsorted_wt_data_i(data)
{
	var len = data.length;
	
	// strip the <pre> & </pre>
	data = data.substr(5, len - 11);
	
	var i;
	var j;
	
	// Pre-create the arrays for word lengths 3-7.
	for (i=3; i<8; i++)
	{
		total_word_lists[i] = [];
	}
	
	var a_lines = data.split("\n");
	var n = a_lines.length;
	
	for (i=0; i<n; i++)
	{
		var a_words = a_lines[i].split(" ");
		var m = a_words.length;
		
		for (j=0; j<m; j++)
		{
			var wd = a_words[j];
			var word_len = wd.length;
			if (word_len < 3 || word_len > 7) continue;
			
			total_word_lists[word_len].push(wd);
			
			var hash = hash_string(wd);
			
			if (total_word_map[hash])
			{
				total_word_map[hash].push(wd);
			}
			else
			{
				total_word_map[hash] = [wd];
			}
		}
	}
}


// load_unsorted_wt_data - load word data from a dictionary that has not been
// pre-sorted by word length.
function load_unsorted_wt_data(frm_name)
{
	var ofrm = $(frm_name);
	var data = ofrm.contentDocument.body.innerHTML;

	load_unsorted_wt_data_i(data);
	
	on_dictionary_frame_loaded(frm_name);
}

// on_dictionary_frame_loaded - determine if all component dictionary files have loaded, and
// if so then update the GUI.
function on_dictionary_frame_loaded(frm_name)
{
	if (!dictionary_load_op)
		return;
		
	var a_frames = dictionary_load_op.a_frames;
	var len = a_frames.length;
	for (var i=0; i<len; i++)
	{
		if (a_frames[i].id == frm_name)
		{
			dictionary_load_op.a_frames.splice(i, 1);
			if (dictionary_load_op.a_frames.length == 0)
			{
				window.setTimeout(dictionary_load_op.cb, 1);
			}
			break;
		}
	}
}


// backup_on_frame_loaded - backup incase frame.onload callback fails to trigger.
function backup_on_frame_loaded(frm_name)
{
	var frm = $(frm_name);
	if (!frm.onload) return;
	
	if (frm.readyState == 4 || frm.readyState == "complete")
	{
		load_unsorted_wt_data(frm_name);
		frm.onload = null;
	}
}


// load_ospd_dictionary - load the OSPD dictionary.
// TODO: this ought to be split into multiple component files as well.
function load_ospd_dictionary()
{
	var frm_name = "ospd";
	var filename = "dict/ospd.txt"

	if (pc_standalone_mode)
		filename = "file:dict/ospd.txt";
		
	reset_dictionary();

	// If we have already loaded this dictionary then use the cached copy.	
	if ($(frm_name))
	{
		load_unsorted_wt_data(frm_name);
		setTimeout(on_load_dictionary_finished, 1);
		return;
	}
	
	// Create a callback function to run when the frame is loaded. But also create a backup timeout as a fail-safe
	// in case the frame.onload callback is never called (I'm looking at you, Microsoft IE).
	function mk_cb() { return function() { return load_unsorted_wt_data(frm_name); $(frm_name).onload = null; } }
	function mk_check_cb() { return function() { backup_on_frame_loaded(frm_name); } }
	
	
	var frm = document.createElement("iframe");
	frm.id = frm_name;
	frm.src = filename;
	frm.onload = mk_cb();
	frm.style.display = "none";
	
/*	var iframe_src = '<iframe id="' + frm_name + '" src="' + filename + '" onload="load_unsorted_wt_data(\'' + frm_name + '\')" display="none" />';
	var spn = document.createElement("span");
	spn.innerHTML = iframe_src;
	document.body.appendChild(spn);
	var frm = $(frm_name); */
	
	// Setup a callback object.
	var cb_obj = new Object();
	cb_obj.a_frames = [ frm ];
	cb_obj.cb = on_load_dictionary_finished;
	dictionary_load_op = cb_obj;
	
	document.body.appendChild(frm);
	
	// Setup a backup callback function in case the regular onload function is never called (IE issue).
	var check_cb = mk_check_cb();
	setTimeout(check_cb, 1000);
}


// load_twlpods_dictionary - load the TWLPODS (intersection of TWL & SOWPODS) dictionary
// from its component files.
function load_twlpods_dictionary()
{
	var a_frames = [];
	
	reset_dictionary();

	for (var n=3; n<=7; n++)
	{
		var frm_name = "w" + n;
		var filename = "dict/" + frm_name + ".txt";
		
		// If we have already loaded this dictionary then use the cached copy.	
		if ($(frm_name))
		{
			load_wt_data(n);
			continue;
		}

		function mk_cb(n) { return function() { return load_wt_data(n); } }
	
		var frm = document.createElement("iframe");
		frm.id = frm_name;
		frm.src = filename;
		frm.onload = mk_cb(n);
		frm.style.display = "none";
		
		a_frames.push(frm);
	}
	
	// Setup a callback object.
	var cb_obj = new Object();
	cb_obj.a_frames = a_frames;
	cb_obj.cb = on_load_dictionary_finished;
	dictionary_load_op = cb_obj;

	for (var i=0; i<a_frames.length; i++)
	{	
		document.body.appendChild(a_frames[i]);
	}
}


// on_load_dictionary_finished - update the GUI after we have finished loading the dictionary.
function on_load_dictionary_finished()
{
	// We keep the start button disabled until the dictionary has finished loading.
	var btn_start = $("btn_start");
	btn_start.disabled = false;
	
	// This is delicate. We really want to update the game parameters (round length, etc)
	// but not clear the revealed word data from the previous round.
	update_game_desc();
	
	dictionary_load_op = null;
}


// load_dictionary - load the whole dictionary into RAM.
// We use iframes by default because Ajax doesn't work in PC standalone mode.
// However, iframe-based loading doesn't work on IE (or at least I haven't figured out how to do it yet).
function load_dictionary(dict_name)
{
	g_dictionary_name = dict_name;
	
	if (dict_name == "ospd")
	{
		if (pc_standalone_mode)
		{
			load_ospd_dictionary();
		}
		else
		{
			ajax_load_ospd_dictionary();
		}
	}
	else if (dict_name == "twlpods")
	{
		load_twlpods_dictionary();
	}
}


// create_loading_indicator - display "Loading..." in the top left corner of the screen.
function create_loading_indicator()
{
	var div = document.createElement("div");
	div.id = "dv_loading";
	div.setAttribute("style","position:absolute;top:0px;left:0px;");
	div.setAttribute("class","loading-indicator");
	div.innerHTML = "Loading...";
	document.body.appendChild(div);
}

function clear_loading_indicator()
{
	var div = $("dv_loading");
	if (div)
		document.body.removeChild(div);
}


function ajax_load_ospd_dictionary()
{
	reset_dictionary();
	
	var filename = "dict/ospd.txt"
	var url = /*g_server_root + */ filename;
	
	create_loading_indicator();
	jx.load(url, on_ajax_load_dictionary_ready, "text", "GET");
}


function on_ajax_load_dictionary_ready(data)
{
	clear_loading_indicator();
dbg_msg("on_ajax_load_dictionary_ready");
	load_unsorted_wt_data_i(data);
	on_load_dictionary_finished();
}



// load_round_from_server - load a single round from the server.
function load_round_from_server(board_num)
{
	create_loading_indicator();

	var url = g_server_root + "cgi-bin/wt_load_round.cgi?bi=" + board_num;
	jx.load(url, load_board_data_i, "text", "GET");
}


// parse_round_header - parse the header for a server round.
function parse_round_header(hdr_string)
{
dbg_msg("hdr_string=" + hdr_string);
	var a_hdr = hdr_string.split(" ");
	if (a_hdr[0] != "#WT1") return false;

	var dict_name = a_hdr[1];
	var bingo_len = parseInt(a_hdr[2]);
	// Note: round time was not configurable in some early versions of header format WT1.	
	var round_time = 60;
	if (a_hdr.length > 2) round_time = parseInt(a_hdr[3]);
	

	if (dict_name != "ospd" && dict_name != "twlpods")
		return false;
		
	if (bingo_len != 6 && bingo_len != 7)
		return false;
	
	g_dictionary_name = dict_name;
	g_bingo_length = bingo_len;
	g_round_time = round_time;
	
	return true;
}


// load_board_data (obsolete) - load a single round in response to data from the server.
// This is the obsolete iframe-based version.
function load_board_data(frm_name)
{
	var ofrm = document.getElementById(frm_name);
	var data = ofrm.contentDocument.body.innerHTML
	load_board_data_i(data);
}

function load_board_data_i(data)
{
	clear_loading_indicator();
	
	var len = data.length;
	
	// Create a new word list for this word length.
	var aWords = []; 

	// Strip any unnecessary HTML tags.
	var hdr1 = "<HTML><PRE>\n";
	var ftr1 = "</PRE></HTML>";
	var hdr2 = "<PRE>";
	var ftr2 = "</PRE>";
	
	
	// TODO: This is added by the CGI script, but we don't really need it.
	if (data.substr(0, hdr1.length) == hdr1)
	{
		data = data.substr(hdr1.length, len - hdr1.length - ftr1.length);
	}
	else if (data.substr(0, hdr2.length) == hdr2)
	{
		// This is for an older version of the script.
		// strip the <pre> & </pre>
		data = data.substr(5, len - 11);
	}
	
	var i;
	var j;
	
	var a_lines = data.split("\n");
	var n = a_lines.length;
	
	// The first line is a header. The second line contains the primary bingo word.
	var hdr_data = a_lines[0];
	if (!parse_round_header(hdr_data))
	{
		dbg_msg("Failed header parsing.");
		return;
	}

	var bingo_word = a_lines[1];
	
	// TODO: currently rather than loading the board data directly, we are loading it into a dictionary.
	for (i=2; i<n; i++)
	{
		var a_words = a_lines[i].split(" ");
		var m = a_words.length;
		
		for (j=0; j<m; j++)
		{
			var wd = a_words[j];
			if (wd.length == 0) continue;
			
			aWords.push(wd);
			
			var hash = hash_string(wd);
			
			if (total_word_map[hash])
			{
				total_word_map[hash].push(wd);
			}
			else
			{
				total_word_map[hash] = [wd];
			}
		}
	}
	
	// Store this word list in RAM.
	total_word_lists[g_bingo_length] = [bingo_word];
	
	on_load_dictionary_finished();
}
